/*
 *  $Id: inode.c,v 1.3 1999/09/15 05:30:59 ezk Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include <fist.h>
#include <wrapfs.h>


STATIC int
wrapfs_create(inode_t *dir, dentry_t *dentry, int mode)
{
    int err = -EACCES;
    dentry_t *hidden_dentry = wrapfs_hidden_dentry(dentry);
    dentry_t *hidden_dir_dentry;

    print_entry_location();
    ASSERT(hidden_dentry != NULL);

    fist_checkinode(dir, "wrapfs_create");

    hidden_dir_dentry = lock_parent(hidden_dentry);
    err = PTR_ERR(hidden_dir_dentry);
    if (IS_ERR(hidden_dir_dentry))
	goto out;
    if (!hidden_dir_dentry->d_inode->i_op || !hidden_dir_dentry->d_inode->i_op->create) {
	unlock_dir(hidden_dir_dentry);
	goto out;
    }
    err = hidden_dir_dentry->d_inode->i_op->create(hidden_dir_dentry->d_inode,
						   hidden_dentry, mode);
    unlock_dir(hidden_dir_dentry);

    if (err)
	goto out;

    err = wrapfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (!err) {
	fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
    }
    fist_checkinode(dir, "post wrapfs_create");

 out:
    print_exit_status(err);
    return err;
}


STATIC dentry_t *
wrapfs_lookup(inode_t *dir, dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dir_dentry = wrapfs_hidden_dentry(dentry->d_parent);
    dentry_t *hidden_dentry;
    const char *name = dentry->d_name.name;

    print_entry_location();

    fist_checkinode(dir, "wrapfs_lookup");

    fist_print_dentry("LOOKUP: dentry IN", dentry);
    fist_print_dentry("LOOKUP: dentry->d_parent IN", dentry->d_parent);
    fist_print_dentry("LOOKUP: hidden_dir_dentry IN", hidden_dir_dentry);
    fist_print_inode("LOOKUP: dir IN", dir);

    if (hidden_dir_dentry->d_inode)
	fist_print_inode("LOOKUP: hidden_dir_dentry->d_inode",
			 hidden_dir_dentry->d_inode);

    /* must initialize dentry operations */
    dentry->d_op = &wrapfs_dops;

    /* increase refcount of base dentry (lookup_dentry will decrement) */
    // THIS IS RIGHT! (don't "fix" it)
    dget(hidden_dir_dentry);

    /* will allocate a new hidden dentry if needed */
    hidden_dentry = lookup_dentry(name, hidden_dir_dentry, 0);
    if (IS_ERR(hidden_dentry)) {
	printk("ERR from hidden_dentry!!!\n");
	err = PTR_ERR(hidden_dentry);
	goto out;
    }

    /* update parent directory's atime */
    fist_copy_attr_atime(dir, hidden_dir_dentry->d_inode);

    /* lookup is special: it needs to handle negative dentries */
    if (!hidden_dentry->d_inode) {
	dtohd(dentry) = hidden_dentry;
	d_add(dentry, NULL);
	fist_print_dentry("lookup hidden", hidden_dentry);
	fist_print_dentry("lookup dentry", dentry);
	// ION: not needed b/c d_inode is already null here
	//	dentry->d_inode = NULL;
	// dput(hidden_dentry);
	goto out;
    }

    fist_dprint(6, "lookup \"%s\" -> inode %d\n", name, hidden_dentry->d_inode->i_ino);
    err = wrapfs_interpose(hidden_dentry, dentry, dir->i_sb, 1);
#if 0
    /*
     * interpose always increments the hidden inode's refcount by one
     * but in this function it was already done by the lookup_dentry
     * above, which eventually calls iget() on hidden_inode, thus
     * increasing the refcount once too many.
     */
    if (!err)
	iput(hidden_dentry->d_inode);
#endif

    fist_checkinode(dentry->d_inode, "wrapfs_lookup OUT: dentry->d_inode:");
    fist_checkinode(dir, "wrapfs_lookup OUT: dir:");

out:
    fist_print_dentry("LOOKUP: dentry->d_parent OUT", dentry->d_parent);
    fist_print_dentry(__FUNCTION__ " OUT hidden_dentry", hidden_dentry);
    fist_print_dentry(__FUNCTION__ " OUT hidden_dir_dentry", hidden_dir_dentry);
    print_exit_status(err);
    return ERR_PTR(err);
}


STATIC int
wrapfs_link(dentry_t *old_dentry, inode_t *dir, dentry_t *new_dentry)
{
    int err = -EPERM;
    dentry_t *hidden_old_dentry = wrapfs_hidden_dentry(old_dentry);
    dentry_t *hidden_new_dentry = wrapfs_hidden_dentry(new_dentry);
    dentry_t *hidden_dir_dentry;

    print_entry_location();

    fist_checkinode(dir, "wrapfs_link-dir");
    fist_checkinode(old_dentry->d_inode, "wrapfs_link-oldinode");

    hidden_dir_dentry = lock_parent(hidden_new_dentry);
    err = PTR_ERR(hidden_dir_dentry);
    if (IS_ERR(hidden_dir_dentry))
	goto out;

    err = -EXDEV;
    if (hidden_dir_dentry->d_inode->i_dev != hidden_old_dentry->d_inode->i_dev)
	goto out_lock;

    err = -EPERM;
    if (!hidden_dir_dentry->d_inode->i_op || !hidden_dir_dentry->d_inode->i_op->link)
	goto out_lock;

    // XXX: DQUOT_INIT(dir->d_inode); // do we need that
    err = hidden_dir_dentry->d_inode->i_op->link(hidden_old_dentry,
						 hidden_dir_dentry->d_inode,
						 hidden_new_dentry);

    if (err)
	goto out_lock;
    err = wrapfs_interpose(hidden_new_dentry, new_dentry, dir->i_sb, 0);
    /*
     * XXX: we have to do a wrapfs_copy_inode (or subset thereof) in
     * *ALL* functions, all file systems, and all OSs!!!
     */
    if (!err)
	fist_copy_attr_timesizes(dir, hidden_new_dentry->d_inode);

 out_lock:
    unlock_dir(hidden_dir_dentry);
 out:
    print_exit_status(err);
    return err;
}


STATIC int
wrapfs_unlink(inode_t *dir, dentry_t *dentry)
{
    int err = -EPERM;
    inode_t *hidden_dir = itohi(dir);
    dentry_t *hidden_dentry = wrapfs_hidden_dentry(dentry);

    print_entry_location();
    ASSERT(hidden_dentry != NULL);

    fist_checkinode(dir, "wrapfs_unlink-dir");
    if (!hidden_dir->i_op || !hidden_dir->i_op->unlink)
	goto out;

    err = hidden_dir->i_op->unlink(hidden_dir, hidden_dentry);
    if (!err)
        fist_copy_attr_times(dir, hidden_dir);
    fist_checkinode(dir, "post wrapfs_unlink-dir");

 out:
    print_exit_status(err);
    return err;
}


STATIC int
wrapfs_symlink(inode_t *dir, dentry_t *dentry, const char *symname)
{
    int err = -EPERM;
    dentry_t *hidden_dentry = wrapfs_hidden_dentry(dentry);
    const char *hidden_symname = symname; /* XXX: can modify this */
    dentry_t *hidden_dir_dentry;

    print_entry_location();

    hidden_dir_dentry = lock_parent(hidden_dentry);
    err = PTR_ERR(hidden_dir_dentry);
    if (IS_ERR(hidden_dir_dentry))
	goto out;

    fist_checkinode(dir, "wrapfs_symlink-dir");
    if (!hidden_dir_dentry->d_inode->i_op ||
	!hidden_dir_dentry->d_inode->i_op->symlink) {
	err = -EPERM;
	goto out_lock;
    }

    err = hidden_dir_dentry->d_inode->i_op->symlink(hidden_dir_dentry->d_inode,
						    hidden_dentry,
						    hidden_symname);
    if (err)
	goto out_lock;
    err = wrapfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (!err)
        fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);

    fist_checkinode(dir, "post wrapfs_symlink-dir");

 out_lock:
    unlock_dir(hidden_dir_dentry);
 out:
    print_exit_status(err);
    return err;
}


STATIC int
wrapfs_mkdir(inode_t *dir, dentry_t *dentry, int mode)
{
    int err = -EPERM;
    dentry_t *hidden_dentry = wrapfs_hidden_dentry(dentry);
    dentry_t *hidden_dir_dentry;

    print_entry_location();

    hidden_dir_dentry = lock_parent(hidden_dentry);
    err = PTR_ERR(hidden_dir_dentry);
    if (IS_ERR(hidden_dir_dentry))
	goto out;

    fist_checkinode(dir, "wrapfs_mkdir-dir");
    if (!hidden_dir_dentry->d_inode->i_op ||
	!hidden_dir_dentry->d_inode->i_op->mkdir) {
	err = -EPERM;
	goto out_lock;
    }

    err = hidden_dir_dentry->d_inode->i_op->mkdir(hidden_dir_dentry->d_inode,
						  hidden_dentry,
						  mode);
    if (err)
	goto out_lock;
    err = wrapfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    if (!err)
        fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);

    fist_checkinode(dir, "post wrapfs_mkdir-dir");

 out_lock:
    unlock_dir(hidden_dir_dentry);
 out:
    print_exit_status(err);
    return err;
}


STATIC int
wrapfs_rmdir(inode_t *dir, dentry_t *dentry)
{
    int err = -EPERM;
    dentry_t *hidden_dentry = wrapfs_hidden_dentry(dentry);
    dentry_t *hidden_dir_dentry;

    print_entry_location();

    hidden_dir_dentry = dget(hidden_dentry->d_parent);
    err = PTR_ERR(hidden_dir_dentry);
    if (IS_ERR(hidden_dir_dentry))
	goto out;

    if (!hidden_dentry->d_inode) {
	err = -ENOENT;
	goto out_dput;
    }

    fist_checkinode(dir, "wrapfs_rmdir-dir");
    if (!hidden_dir_dentry->d_inode->i_op ||
	!hidden_dir_dentry->d_inode->i_op->rmdir) {
	err = -EPERM;
	goto out_dput;
    }

    /*
     * The dentry->d_count stuff confuses d_delete() enough to
     * not kill the inode from under us while it is locked. This
     * wouldn't be needed, except the dentry semaphore is really
     * in the inode, not in the dentry..
     */
    hidden_dentry->d_count++;
    double_lock(hidden_dir_dentry, hidden_dentry);

    err = vfs_rmdir(hidden_dir_dentry->d_inode, hidden_dentry);
    if (!err)
        fist_copy_attr_times(dir, hidden_dir_dentry->d_inode);

    double_unlock(hidden_dentry, hidden_dir_dentry);

 out_dput:
    /* don't dput here, it will be done later in d_delete */
    //    dput(hidden_dentry);
 out:
    print_exit_status(err);
    return err;
}


STATIC int
wrapfs_mknod(inode_t *dir, dentry_t *dentry, int mode, int dev)
{
    int err = -EPERM;
    dentry_t *hidden_dentry = wrapfs_hidden_dentry(dentry);
    dentry_t *hidden_dir_dentry;

    print_entry_location();

    hidden_dir_dentry = lock_parent(hidden_dentry);
    err = PTR_ERR(hidden_dir_dentry);
    if (IS_ERR(hidden_dir_dentry))
	goto out;

    fist_checkinode(dir, "wrapfs_mknod-dir");
    if (!hidden_dir_dentry->d_inode->i_op || !hidden_dir_dentry->d_inode->i_op->mknod) {
	err = -EPERM;
	goto out_lock;
    }

    err = hidden_dir_dentry->d_inode->i_op->mknod(hidden_dir_dentry->d_inode,
						  hidden_dentry,
						  mode,
						  dev);
    if (err)
	goto out_lock;
    err = wrapfs_interpose(hidden_dentry, dentry, dir->i_sb, 0);
    fist_checkinode(dir, "post wrapfs_mknod-dir");
    if (!err)
        fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);

 out_lock:
    unlock_dir(hidden_dir_dentry);
 out:
    print_exit_status(err);
    return err;
}


STATIC int
wrapfs_rename(inode_t *old_dir, dentry_t *old_dentry,
	      inode_t *new_dir, dentry_t *new_dentry)
{
    int err = -EPERM;
    dentry_t *hidden_old_dentry = wrapfs_hidden_dentry(old_dentry);
    dentry_t *hidden_new_dentry = wrapfs_hidden_dentry(new_dentry);
    dentry_t *hidden_old_dir_dentry;
    dentry_t *hidden_new_dir_dentry;

    print_entry_location();

    hidden_old_dir_dentry = get_parent(hidden_old_dentry);
    hidden_new_dir_dentry = get_parent(hidden_new_dentry);
    err = PTR_ERR(hidden_old_dir_dentry);
    if (IS_ERR(hidden_old_dir_dentry))
	goto out;
    err = PTR_ERR(hidden_new_dir_dentry);
    if (IS_ERR(hidden_new_dir_dentry))
	goto out;
    double_lock(hidden_old_dir_dentry, hidden_new_dir_dentry);

    fist_checkinode(old_dir, "wrapfs_rename-old_dir");
    fist_checkinode(new_dir, "wrapfs_rename-new_dir");

    err = -ENOENT;
    if (check_parent(hidden_old_dir_dentry, hidden_old_dentry) &&
	check_parent(hidden_new_dir_dentry, hidden_new_dentry))
	err = vfs_rename(hidden_old_dir_dentry->d_inode, hidden_old_dentry,
			 hidden_new_dir_dentry->d_inode, hidden_new_dentry);
    if (err)
	goto out_lock;

    if (!err) {
        fist_copy_attr_all(new_dir, hidden_new_dir_dentry->d_inode);
	if (new_dir != old_dir)
	    fist_copy_attr_times(old_dir, hidden_old_dir_dentry->d_inode);
    }

    fist_checkinode(new_dir, "post wrapfs_mknod-new_dir");

 out_lock:
    double_unlock(hidden_old_dir_dentry, hidden_new_dir_dentry);
    // new do_rename has dput() after double_unlock.  Is it needed?
 out:
    print_exit_status(err);
    return err;
}


STATIC int
wrapfs_readlink(dentry_t *dentry, char *buf, int bufsiz)
{
    int err;
    dentry_t *hidden_dentry = wrapfs_hidden_dentry(dentry);

    print_entry_location();
    fist_print_dentry("wrapfs_readlink IN", dentry);

    if (!hidden_dentry->d_inode->i_op || !hidden_dentry->d_inode->i_op->readlink) {
	err = -EPERM;
	goto out;
    }

    err = hidden_dentry->d_inode->i_op->readlink(hidden_dentry,
						 buf,
						 bufsiz);
    if (err > 0)
	fist_copy_attr_atime(dentry->d_inode, hidden_dentry->d_inode);

 out:
    print_exit_status(err);
    return err;
}


STATIC dentry_t *
wrapfs_follow_link(dentry_t *dentry, dentry_t *base, unsigned int follow)
{
    char *buf;
    int len = PAGE_SIZE, err;
    dentry_t *ret_dentry = dentry;
    dentry_t *hidden_dentry = wrapfs_hidden_dentry(dentry);
    mm_segment_t old_fs;
    //    dentry_t *hidden_base = wrapfs_hidden_dentry(base);

    print_entry_location();
    fist_print_dentry("wrapfs_follow_link dentry IN", dentry);
    fist_print_dentry("wrapfs_follow_link base IN", base);

    if (!hidden_dentry->d_inode->i_op ||
	!hidden_dentry->d_inode->i_op->readlink ||
	!hidden_dentry->d_inode->i_op->follow_link) {
	/*
	 * do_follow_link will do a dput() on the dentry passed to us
	 * and if that is the same as ret_dentry, and we want it kept,
	 * then we must increment the dentry's refcount to avoid having
	 * it deallocate/destroyed.
	 */
	dget(ret_dentry);
	/*
	 * Looking at do_follow_link, if we implement follow_link, then
	 * we have to dput(base).  Normally that's done by lookup_dentry
	 * but here we don't run that code.
	 */
	dput(base);
	goto out;
    }

    buf = kmalloc(len, GFP_KERNEL);
    if (!buf) {
	ret_dentry = ERR_PTR(-ENOMEM);
	dput(base); // for the same reason as above
	goto out;
    }
    /* read the hidden symlink, and then we will follow it */
    old_fs = get_fs();
    set_fs(KERNEL_DS);
    err = hidden_dentry->d_inode->i_op->readlink(hidden_dentry, buf, len);
    set_fs(old_fs);
    if (err < 0) {
	ret_dentry = ERR_PTR(err);
	dput(base); // for the same reason as above
	goto out_free;
    }
    fist_copy_attr_atime(dentry->d_inode, hidden_dentry->d_inode);
    buf[err] = 0;	// terminate the buffer for lookup_dentry's sake

#if 0
    /* run dget b/c lookup_dentry expects the caller to do so */
    dget(base);
#endif
    /*
     * We do not dget(base) b/c the way we were called here,
     * the base (passed to us) was already dget'ed.
     */
    ret_dentry = lookup_dentry(buf, base, follow);
    fist_print_dentry("wrapfs_follow_link base OUT", base);

 out_free:
    kfree_s(buf, len);
 out:
    print_exit_location();
    return ret_dentry;
}


STATIC int
wrapfs_get_block(inode_t *inode, long iblock, struct buffer_head *bh_result, int create)
{
    int err = -EINVAL; // XXX/ION/EZK: NOT IMPLEMENTED YET
    inode_t *hidden_inode = itohi(inode);

    panic("YOU SHOULD IMPLEMENT GET_BLOCK FIRST\n");

    if (hidden_inode->i_op && hidden_inode->i_op->get_block)
	err = hidden_inode->i_op->get_block(hidden_inode,
					    iblock, bh_result, create);

    print_exit_status(err);
    return err;
}


STATIC int
wrapfs_readpage(file_t *file, page_t *page)
{
    int err = 0;
    inode_t *inode = file->f_dentry->d_inode;
    inode_t *hidden_inode = itohi(inode);
    file_t *hidden_file = ftohf(file);
    page_t *hidden_page, **hash;
    unsigned long page_cache = 0, page_data, hidden_page_data;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,16)
    int free;
#endif /* kernel before 2.3.16 */

    print_entry_location();

    ASSERT(hidden_inode->i_op != NULL);
    fist_dprint(6, "READPAGE: f_pos:0x%x inode:0x%x i_size:0x%x hidden_inode:0x%x page->inode:0x%x page->offset:%lu\n",
		(int) file->f_pos, (int) inode, inode->i_size, (int) hidden_inode,
		(int) page->inode, page->offset);

    /* create new hidden_page (look for it in cache?) */
    hash = page_hash(hidden_inode, page->offset & PAGE_MASK);

restart_page:
    // must lock hidden pages in 2.3
    hidden_page = find_lock_page(hidden_inode, page->offset & PAGE_MASK);
    if (!hidden_page) {
	if (!page_cache) {		/* allocate new page */
	    page_cache = __get_free_page(GFP_KERNEL);
	    if (!page_cache) {
		err = -ENOMEM;
		goto out;		/* no need to free anything */
	    }
	    goto restart_page;
	}
	hidden_page = mem_map + MAP_NR(page_cache);
	page_cache = 0;
	add_to_page_cache(hidden_page, hidden_inode, page->offset & PAGE_MASK);
    }

    /* now we have a valid hidden page */

    fist_dprint(6, "READPAGE: hidden_page %x\n", (int) hidden_page);

    /*
     * XXX: some times the hidden_page shows up as up-to-date (?!)
     * If we call readpage on lower f/s, we get a BUG() in fs/buffer.c,
     * line 1669 in block_read_full_page().  So we turn off the bit.
     * We should do something better and more optimized in here, such as
     * skip calling the lower f/s if page is up-to-date.
     * -Erez
     */
    ClearPageUptodate(hidden_page);
    ASSERT(!Page_Uptodate(hidden_page));

    if (hidden_inode->i_op->readpage) {
	/*
	 * create new hidden_page (look for it in cache?)
	 * call lower level readpage on hidden_inode and hidden_page
	 */
	// ION, I add this. needed?
	file->f_pos = hidden_file->f_pos = page->offset;

#if 0
	hidden_file->f_reada = file->f_reada;
	hidden_file->f_ramax = file->f_ramax;
	hidden_file->f_raend = file->f_raend;
	hidden_file->f_ralen = file->f_ralen;
	hidden_file->f_rawin = file->f_rawin;
#endif

	err = hidden_inode->i_op->readpage(hidden_file, hidden_page);
	if (err) {
	    printk("READPAGE: error reading hidden page: %d\n", err);
	    goto out_free;
	}

	/* XXX: may need to wait_on_page(hidden_page) */
	wait_on_page(hidden_page);
	if (!Page_Uptodate(hidden_page) || PageError(hidden_page)) {
	    printk("hidden page not up-to-date or error!\n");
	    err = -EIO;
	    goto out_free;
	}
    }
    /* XXX: what if the lower f/s doesn't implement readpage? */

    /*
     * decode hidden_page data onto page
     */
    page_data = page_address(page);
    hidden_page_data = page_address(hidden_page);

    wrapfs_decode_block((char *) hidden_page_data, (char *) page_data, PAGE_SIZE);

    /*
     * adjust flags and wake up processes waiting on the page
     * code shamelessly stolen from generic_readpage
     */
    // ION, why? we never set the locked bit?
    clear_bit(PG_locked, &page->flags);
    set_bit(PG_uptodate, &page->flags);
    wake_up(&page->wait);

    // ION/EZK: this code taken from 2.3.9-pre7's fs/buffer.c's
    // end_buffer_io_async() function
    /*
     * Note - we need to test the flags before we unlock the page, but
     * we must not actually free the page until after the unlock!
     */
    if (test_and_clear_bit(PG_decr_after, &page->flags))
	atomic_dec(&nr_async_pages);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,16)
    if (test_and_clear_bit(PG_free_swap_after, &page->flags))
	swap_free(page->offset);
    free = test_and_clear_bit(PG_free_after, &page->flags);
#endif /* kernel before 2.3.16 */

#if 0
# error THIS CODE IS NEEDED!!!?
    // is this needed?
    if (page->owner != -1)
	PAGE_BUG(page);
    page->owner = (int)current;
    UnlockPage(page);

    if (free)
	__free_page(page);
#else
# if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,16)
    ASSERT(free == 0);		/* XXX: tell me something for now */
# endif /* kernel before 2.3.16 */
#endif

out_free:
    if (page_cache) {
	printk("READPAGE: page cache is not null: 0x%x\n", (int) page_cache);
	free_page(page_cache);
    }
#if 0
    if (hidden_page->owner != -1)
	PAGE_BUG(hidden_page);
    hidden_page->owner = (int)current;
    UnlockPage(hidden_page);	/* EZK/ION:XXX: is this right? */
#endif
    // ION, do we need to __free_page? isn't is sub-optimal
    __free_page(hidden_page);	/* release_page() in mm/filemap.c */
out:
    fist_print_page_flags("READPAGE flags (exit)", page);
    print_exit_status(err);
    return err;
}


/* XXX: the VFS never calls writepage */
STATIC int
wrapfs_writepage(file_t *file, page_t *page)
{
    int err = -EIO;
    file_t *hidden_file = ftohf(file);
    inode_t *hidden_inode = hidden_file->f_dentry->d_inode;

    panic("ARE YOU SURE YOU WANT TO DEFINE WRITEPAGE???\n");
    print_entry_location();

    if (hidden_inode->i_op && hidden_inode->i_op->writepage) {
	fist_dprint(2, "non-generic writepage\n");
	err = hidden_inode->i_op->writepage(hidden_file, page);
    }

    print_exit_status(err);
    return err;
}


// new in 2.3
// ION/EZK: is this code correct?
STATIC int
wrapfs_flushpage(inode_t *inode, page_t *page, unsigned long offset)
{
    int err = -EINVAL;
    inode_t *hidden_inode = itohi(inode);
    page_t *hidden_page = NULL;

    print_entry_location();

    ASSERT(page);
#if 0
    if (page->offset != offset) {
	printk(KERN_WARNING "FLUSHPAGE: mismatching offsets 0x%x 0x%x\n",
	       page->offset, offset);
	err = -EIO;
	goto out;
    }
#endif

    if (!hidden_inode->i_op || !hidden_inode->i_op->flushpage)
	goto out;

    hidden_page = find_lock_page(hidden_inode, page->offset);
    if (!hidden_page) {
	printk(KERN_ERR "flushpage: cannot find hidden_page at offset 0x%x\n",
	       (int) page->offset);
	err = -EIO;
	goto out;
    }

    if (!PageLocked(hidden_page))
	PAGE_BUG(hidden_page);

    err = hidden_inode->i_op->flushpage(hidden_inode, hidden_page, hidden_page->offset);

    if (!PageLocked(hidden_page))
	PAGE_BUG(hidden_page);

#if 0
    SetPageUptodate(hidden_page);
#else
    ClearPageUptodate(hidden_page);
    lru_cache_del(hidden_page);
    remove_inode_page(hidden_page);
    lru_cache_del(page);
#endif
    // no need to UnlockPage(page) b/c VFS un/locks it
    UnlockPage(hidden_page);
    page_cache_release(hidden_page);
    page_cache_release(hidden_page);

out:
    print_exit_status(err);
    return err;
}


/* kludgey - replace when up/down problem fixed */
STATIC void
wrapfs_truncate(inode_t *inode)
{
    inode_t *hidden_inode = itohi(inode);

    print_entry_location();

    fist_checkinode(inode, "wrapfs_truncate");
    if (!hidden_inode->i_op || !hidden_inode->i_op->truncate)
	goto out;

    down(&hidden_inode->i_sem);
    hidden_inode->i_op->truncate(hidden_inode);
    up(&hidden_inode->i_sem);

    fist_copy_attr_timesizes(inode, hidden_inode);

 out:
    print_exit_location();
}


STATIC int
wrapfs_permission(inode_t *inode, int mask)
{
    inode_t *hidden_inode = itohi(inode);
    int mode = hidden_inode->i_mode;
    int err;

    print_entry_location();
    //    err = -ENOSPC;goto out;
    if (hidden_inode->i_op && hidden_inode->i_op->permission) {
	err = hidden_inode->i_op->permission(hidden_inode, mask);
	goto out;
    }

    if ((mask & S_IWOTH) && IS_RDONLY(hidden_inode) &&
	(S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode))) {
	err = -EROFS; /* Nobody gets write access to a read-only fs */
	goto out;
    }
    if ((mask & S_IWOTH) && IS_IMMUTABLE(hidden_inode)) {
	err = -EACCES; /* Nobody gets write access to an immutable file */
	goto out;
    }
    if (current->fsuid == hidden_inode->i_uid)
	mode >>= 6;
    else if (in_group_p(hidden_inode->i_gid))
	mode >>= 3;

    if (((mode & mask & S_IRWXO) == mask) || capable(CAP_DAC_OVERRIDE)) {
	err = 0;
	goto out;
    }

    /* read and search access */
    if ((mask == S_IROTH) ||
	(S_ISDIR(mode)  && !(mask & ~(S_IROTH | S_IXOTH))))
	if (capable(CAP_DAC_READ_SEARCH)) {
	    err = 0;
	    goto out;
	}
    err = -EACCES;

 out:
    print_exit_status(err);
    return err;
}


STATIC int
wrapfs_smap(inode_t *inode, int block)
{
    int err = -EINVAL;
    inode_t *hidden_inode = itohi(inode);

    print_entry_location();

    if (hidden_inode->i_op && hidden_inode->i_op->smap)
	err = hidden_inode->i_op->smap(hidden_inode, block);
    else
	err = bmap(hidden_inode, block);

    print_exit_status(err);
    return err;
}


STATIC int
wrapfs_inode_revalidate(dentry_t *dentry)
{
    int err = 0;
    dentry_t *hidden_dentry = wrapfs_hidden_dentry(dentry);
    inode_t *hidden_inode = hidden_dentry->d_inode;

    print_entry_location();
    fist_print_dentry("inode_revalidate IN", dentry);

    if (hidden_inode->i_op && hidden_inode->i_op->revalidate) {
	err = hidden_inode->i_op->revalidate(hidden_dentry);
	if (!err)
	    fist_copy_attr_all(dentry->d_inode, hidden_inode);
    }

    fist_print_dentry("inode_revalidate OUT", dentry);
    print_exit_status(err);
    return err;
}


struct inode_operations wrapfs_iops =
{
    &wrapfs_fops,		/* default file operations */
    wrapfs_create,		/* create */
    wrapfs_lookup,		/* lookup */
    wrapfs_link,		/* link */
    wrapfs_unlink,		/* unlink */
    wrapfs_symlink,		/* symlink */
    wrapfs_mkdir,		/* mkdir */
    wrapfs_rmdir,		/* rmdir */
    wrapfs_mknod,		/* mknod */
    wrapfs_rename,		/* rename */
    wrapfs_readlink,		/* readlink */
#if 1
    wrapfs_follow_link,		/* follow_link */
#else
    NULL,			/* follow link */
#endif
    wrapfs_get_block,		/* get_block */
    wrapfs_readpage,		/* readpage */
    wrapfs_writepage,		/* writepage */
    wrapfs_flushpage,		/* flushpage */
    wrapfs_truncate,		/* truncate */
    wrapfs_permission,		/* permission */
    wrapfs_smap,		/* smap */
    wrapfs_inode_revalidate	/* revalidate */
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
